/*
    Trackable API

function CreateTrackable takes string s, real x, real y, real f returns trackable
function CreateTrackableLoc takes string s, real x, real y, real f returns trackable
function CreateTrackableForPlayer takes player whichPlayer, string s, real x, real y, real f returns trackable
function CreateTrackableLocForPlayer takes player whichPlayer, string s, location loc, real f returns trackable
function FlushTrackableData takes trackable t returns nothing

function GetTrackableX takes trackable t returns real
function GetTrackableY takes trackable t returns real
function GetTrackableZ takes trackable t returns real
function GetTrackableFacing takes trackable t returns real
function GetTrackableModelPath takes trackable t returns string
function IsTrackableVisibleForPlayer takes trackable t, player whichPlayer returns boolean

function FNWETriggerRegisterAnyTrackableHitEvent takes trigger trig returns nothing
function FNWETriggerRegisterAnyTrackableTrackEvent takes trigger trig returns nothing


*/

#guard fn_trackableAPI

library FNWETrackableAPI initializer InitAnyHitTrackEventSystem{
    trigger fn_triggerRegisterAnyTrackableHitEvent = CreateTrigger()
    trigger fn_triggerRegisterAnyTrackableTrackEvent = CreateTrigger()
    
    private trigger array AnyTrackableHitEvent
    private trigger array AnyTrackableTrackEvent
    
    private integer fn_triggerHitCounter = 0
    private integer fn_triggerTrackCounter = 0

define{
    <call CreateTrackable> (s,x,y,f) = CreateTrackable_hook(s,x,y,f)
    <CreateTrackable> (s,x,y,f) = CreateTrackable_hook(s,x,y,f)
    
    private FNWE_HASHTABLE_MODEL_PATH_KEY = 0
    private FNWE_HASHTABLE_X_KEY = 1
    private FNWE_HASHTABLE_Y_KEY = 2
    private FNWE_HASHTABLE_FACING_KEY = 3
    private FNWE_HASHTABLE_PLAYER_INDEX_KEY = 4
}

trackable CreateTrackable_hook(string s, real x, real y, real f){
    trackable t = Create##Trackable(s,x,y,f)
    integer h = GetHandleId(t)
    SaveStr(FNWEHashtable,h,FNWE_HASHTABLE_MODEL_PATH_KEY,s)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_X_KEY,x)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_Y_KEY,y)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_FACING_KEY,f)
    SaveInteger(FNWEHashtable,h,FNWE_HASHTABLE_PLAYER_INDEX_KEY,-1)
    TriggerRegisterTrackableHitEvent(fn_triggerRegisterAnyTrackableHitEvent,t)
    TriggerRegisterTrackableTrackEvent(fn_triggerRegisterAnyTrackableTrackEvent,t)
    return t
}

trackable CreateTrackableLoc(string s, location loc, real f){
    real x = GetLocationX(loc)
    real y = GetLocationY(loc)
    trackable t = Create##Trackable(s,x,y,f)
    integer h = GetHandleId(t)
    SaveStr(FNWEHashtable,h,FNWE_HASHTABLE_MODEL_PATH_KEY,s)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_X_KEY,x)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_Y_KEY,y)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_FACING_KEY,f)
    SaveInteger(FNWEHashtable,h,FNWE_HASHTABLE_PLAYER_INDEX_KEY,-1)
    TriggerRegisterTrackableHitEvent(fn_triggerRegisterAnyTrackableHitEvent,t)
    TriggerRegisterTrackableTrackEvent(fn_triggerRegisterAnyTrackableTrackEvent,t)
    loc = null
    return t
}

trackable CreateTrackableForPlayer(player whichPlayer, string s, real x, real y, real f){
    string model = ""
    trackable t = null
    integer h = 0
    if GetLocalPlayer() == whichPlayer{
        model = s
    }
    t = Create##Trackable(model,x,y,f)
    h = GetHandleId(t)
    SaveStr(FNWEHashtable,h,FNWE_HASHTABLE_MODEL_PATH_KEY,s)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_X_KEY,x)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_Y_KEY,y)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_FACING_KEY,f)
    SaveInteger(FNWEHashtable,h,FNWE_HASHTABLE_PLAYER_INDEX_KEY,GetPlayerId(whichPlayer))
    TriggerRegisterTrackableHitEvent(fn_triggerRegisterAnyTrackableHitEvent,t)
    TriggerRegisterTrackableTrackEvent(fn_triggerRegisterAnyTrackableTrackEvent,t)
    whichPlayer = null
    return t
}

trackable CreateTrackableLocForPlayer(player whichPlayer, string s, location loc, real f){
    string model = ""
    real x = GetLocationX(loc)
    real y = GetLocationY(loc)
    trackable t = null
    integer h = 0
    if GetLocalPlayer() == whichPlayer{
        model = s
    }
    t = Create##Trackable(model,x,y,f)
    h = GetHandleId(t)
    SaveStr(FNWEHashtable,h,FNWE_HASHTABLE_MODEL_PATH_KEY,s)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_X_KEY,x)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_Y_KEY,y)
    SaveReal(FNWEHashtable,h,FNWE_HASHTABLE_FACING_KEY,f)
    SaveInteger(FNWEHashtable,h,FNWE_HASHTABLE_PLAYER_INDEX_KEY,GetPlayerId(whichPlayer))
    TriggerRegisterTrackableHitEvent(fn_triggerRegisterAnyTrackableHitEvent,t)
    TriggerRegisterTrackableTrackEvent(fn_triggerRegisterAnyTrackableTrackEvent,t)
    loc = null
    whichPlayer = null
    return t
}

real GetTrackableX(trackable t){
    return LoadReal(FNWEHashtable,GetHandleId(t),FNWE_HASHTABLE_X_KEY)
}

real GetTrackableY(trackable t){
    return LoadReal(FNWEHashtable,GetHandleId(t),FNWE_HASHTABLE_Y_KEY)
}

real GetTrackableZ(trackable t){
    integer h = GetHandleId(t)
    location loc = Location(LoadReal(FNWEHashtable,h,FNWE_HASHTABLE_X_KEY),LoadReal(FNWEHashtable,h,FNWE_HASHTABLE_Y_KEY))
    real z = GetLocationZ(loc)
    RemoveLocation(loc)
    loc = null
    t = null
    return z
}

real GetTrackableFacing(trackable t){
    return LoadReal(FNWEHashtable,GetHandleId(t),FNWE_HASHTABLE_FACING_KEY)
}

string GetTrackableModelPath(trackable t){
    return LoadStr(FNWEHashtable,GetHandleId(t),FNWE_HASHTABLE_MODEL_PATH_KEY)
}

boolean IsTrackableVisibleForPlayer(trackable t, player whichPlayer){
    integer i = LoadInteger(FNWEHashtable,GetHandleId(t),FNWE_HASHTABLE_PLAYER_INDEX_KEY)
    if i == 1{
        whichPlayer = null
        return true
    }else{
        if GetPlayerId(whichPlayer) == i{
            return true
        }else{
            return false
        }
    }
}

private nothing AnyHitEventAction(){
    integer i = 0
    loop{
        exitwhen i == 8192 or AnyTrackableHitEvent[i] == null
        if TriggerEvaluate(AnyTrackableHitEvent[i]){
            TriggerExecute(AnyTrackableHitEvent[i])
        }
        i++
    }
}

private nothing AnyTrackEventAction(){
    integer i = 0
    loop{
        exitwhen i == 8192 or AnyTrackableTrackEvent[i] == null
        if TriggerEvaluate(AnyTrackableTrackEvent[i]){
            TriggerExecute(AnyTrackableTrackEvent[i])
        }
        i++
    }
}

nothing FNWETriggerRegisterAnyTrackableHitEvent(trigger trig){
    if trig == null{
        return
    }
    AnyTrackableHitEvent[fn_triggerHitCounter] = trig
    fn_triggerHitCounter++
}

nothing FNWETriggerRegisterAnyTrackableTrackEvent(trigger trig){
    if trig == null{
        return
    }
    AnyTrackableTrackEvent[fn_triggerTrackCounter] = trig
    fn_triggerTrackCounter++
}

public nothing InitAnyHitTrackEventSystem(){
    TriggerAddAction(fn_triggerRegisterAnyTrackableHitEvent,function AnyHitEventAction)
    TriggerAddAction(fn_triggerRegisterAnyTrackableTrackEvent,function AnyTrackEventAction)
}

nothing FlushTrackableData(trackable t){
    integer h = GetHandleId(t)
    t = null
    RemoveSavedString(FNWEHashtable,h,FNWE_HASHTABLE_MODEL_PATH_KEY)
    RemoveSavedReal(FNWEHashtable,h,FNWE_HASHTABLE_X_KEY)
    RemoveSavedReal(FNWEHashtable,h,FNWE_HASHTABLE_Y_KEY)
    RemoveSavedReal(FNWEHashtable,h,FNWE_HASHTABLE_FACING_KEY)
    RemoveSavedInteger(FNWEHashtable,h,FNWE_HASHTABLE_PLAYER_INDEX_KEY)
}

}